"""Module defining an environment (ie folder) that a task needs to run"""
# pylint: disable=too-few-public-methods
import glob
from pathlib import Path
import shutil
import os
from functools import wraps

def run_only_on_root(func):
    """Only runs func on the root node"""
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        if self.on_root:
            func(self, *args, **kwargs)
    return wrapped

class Envs:
    """Collection of all enviroments"""
    def __init__(self, worker_id, on_root):
        """Initialises a new set of environments"""
        self.envs = dict()
        self.worker_id = worker_id
        self.on_root = on_root
        self.root_dir = Path(".").resolve() # Save root folder from first call

    def get_env(self, env_type, mpi_comm):
        """
        Moves the CWD to the specified environment.
        Creates the environment if it doesn't already exist
        """
        if not env_type in self.envs:
            self.envs[env_type] = env_type(self.root_dir, self.worker_id, self.on_root).create()
            mpi_comm.barrier()
        self.envs[env_type].activate()

class Env:
    """Represents a working environment for a task"""
    def __init__(self, root_dir: Path, worker_id, on_root, working_dir = "work", **kwargs):
        """Sets up a environment for worker to work in"""
        # Store main paths in worker
        self.root = root_dir
        self.on_root = on_root
        self.worker_id = worker_id
        self.global_working_dir: Path = self.root / working_dir
        self.copy_list = []
        self.id_name = None

        # Run user provided initialisation with optional kwargs
        self.setup(**kwargs)
        self.local_working_dir = self.global_working_dir / f"worker-{worker_id}-{self.id_name}"

    def create(self):
        """Build actual local folder"""
        self._reset_folder()
        return self

    @run_only_on_root
    def _reset_folder(self):
        """Resets the environment to brand new"""
        self._make_working_dir(delete_old=False)
        if self.copy_list:
            self._do_copy()

    def setup(self):
        """
        Method that can be replaced by concrete workers.
        Should provide a unique name for the environment type.
        """
        raise NotImplementedError("Should override env setup to provide a name!")


    def activate(self):
        """Move cwd to working directory"""
        os.chdir(self.local_working_dir)

    @run_only_on_root
    def _make_working_dir(self, delete_old=True):
        """Creates the working directory for the task"""
        # Delete the old working dir if requested
        if delete_old and self.local_working_dir.exists():
            shutil.rmtree(self.local_working_dir)

        # Make the new folder ignoring whether it exists
        self.local_working_dir.mkdir(exist_ok=True, parents=True)

    @run_only_on_root
    def _do_copy(self):
        """Actually does the copy"""
        for file in self.copy_list:
            filename = file.name
            shutil.copy(file, self.local_working_dir / filename)

    @run_only_on_root
    def copy_files(self, copy_dir=None, copy_glob=None, copy_list=None):
        """Copies files from copy_dir to the working directory.
           You can specify the files to copy as either:
           copy_dir  - copy all the files in the folder
                       e.g. copy_file(copy_dir="myfiles")
           copy_glob - copy files matching a unix glob
                       e.g. copy_file(copy_glob="myfiles/proto.*")
           copy_list - copy all the files in the list
                       e.g. copy_file(copy_list=["myfiles/proto.param","myfiles/proto.cell"])
        """
        # Only root node does copying
        if not self.on_root:
            return

        if copy_dir:
            folder = self.root / copy_dir
            from_files = folder.iterdir()
        elif copy_glob:
            from_files = glob.glob(self.root / copy_glob)
        elif copy_list:
            from_files = map(lambda x: self.root/x, copy_list)
        else:
            raise ValueError("No items provided to copy")

        # Store these to be copied in a moment
        self.copy_list += list(from_files)



"""Example of simple generic tasks"""
#pylint: disable=unused-argument
from task import Task, requires_env
from worker_environment import Env

class GenericEnv(Env):
    """Example of a generic env that is just an empty folder"""
    def setup(self):
        self.id_name = "generic"


class GenericTask(Task):
    """Example of a generic task type"""

    @staticmethod                # Define as a static method (doesn't receive self)
    @requires_env(GenericEnv)    # Decorator to say we want to be in GenericEnv to run this
    def return_zero(mpi_comm):
        """Task that does nothing but return 0. Shows how task functions can be implemented"""
        return 0                 # Return the result that you want to send back to master

    @staticmethod
    @requires_env(GenericEnv)
    def add(mpi_comm, val1, val2):
        """Task that adds two numbers"""
        return val1 + val2



"""Results class to hold results from calculations in a standard way"""
from . import units

default_energy_unit = units.hartree
default_force_unit = units.hartree / units.bohr

class Results:
    def __init__(self, energy=None, forces=None, stress=None, free_energy=None):
        self.energy = energy
        self.forces = forces
        self.stress = stress
        self.free_energy=free_energy


    def __repr__(self):
        return f"energy={self.energy}\tforces={self.forces}"



"""Module containing CASTEP specific task information"""
#pylint: disable=import-outside-toplevel, no-self-argument
from pathlib import Path
import sys
import os
import numpy as np
from ase import Atoms
import ase.io
from .task import Task, requires_env
from .worker_environment import Env
from ..utils.results import Results
from ..utils import units

ATOMS = {}
PARAMS = {}
RESULTS = {}
SEEDNAME = "run"

class CastepDebugEnv(Env):
    """Get a folder to run this instance of CASTEP in"""
    def setup(self):
        self.id_name = "castep-debug"


def write_dict(filename, options):
    """Write options to file"""
    with open(filename, "w") as file_handle:
        for key, val in options.items():
            file_handle.write(f"{key} : {val}\n")


class CastepDebugTask(Task):
    """CASTEP specific tasks -- This version runs in serial by launching the executable manually"""

    @staticmethod 
    def get_default_parameters(task, atoms, **extra_args):
        """Return a default set of parameters for this task -- can be called on master"""
        if task in ["energy", "singlepoint", "SPE", "forces"]:
            params = {
                "cutoff_energy": "400 eV",
                "task": "SPE"
            }
        elif task in ['geometry']:
            params = {
                "cutoff_energy": "400 eV",
                "task": "geometryoptimisation"
            }
        else:
            raise ValueError("Invalid value for task: {}".format(task))

        # Now use the override values the user provided
        params.update(extra_args)
        return params

    @staticmethod
    @requires_env(CastepDebugEnv)
    def init(mpi_comm, atoms, params):
        """Initialise the cell and parameter data -- also initialise DFT code here if possible"""
        global ATOMS, PARAMS
        ATOMS = atoms
        PARAMS = params

    @staticmethod
    @requires_env(CastepDebugEnv)
    def finalise(mpi_comm):
        """Delete any static variables -- also call code cleanup methods here"""
        global ATOMS, PARAMS, RESULTS
        ATOMS = {}
        PARAMS = {}
        RESULTS = {}

    @staticmethod
    @requires_env(CastepDebugEnv)
    def update_params(mpi_comm, new_params):
        """Update the parameters set"""
        # Do some validation checking here
        pass

        # Now update the saved params
        PARAMS.update(new_params)

    @staticmethod
    @requires_env(CastepDebugEnv)
    def update_atoms(mpi_comm, new_atoms):
        """Update the atoms list"""
        global ATOMS

        # Optional validation checking
        pass

        # Now update saved atoms
        ATOMS = new_atoms

    @staticmethod
    @requires_env(CastepDebugEnv)
    def calculate(mpi_comm):
        """Do the actual calculation using the saved params and atoms"""
        global RESULTS

        param_file = f"{SEEDNAME}.param"
        cell_file = f"{SEEDNAME}.cell"

        if mpi_comm.Get_rank() == 0:
            # Write cell with ASE
            ase.io.write(cell_file,ATOMS)

            # Write params file
            write_dict(param_file, PARAMS)

        # Now run the actual code
        if mpi_comm.Get_rank() == 0:

            # Hacky way to launch mpirun version of process
            import subprocess
            from ase.calculators.castep import Castep

            castep_ase = Castep(directory=".", label=SEEDNAME)
            castep_filename = f"{SEEDNAME}.castep"
            if os.path.exists(castep_filename):
                os.remove(castep_filename)

            process_cmd = "castep.serial", SEEDNAME
            process = subprocess.run(process_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            #print(process.stdout, file=sys.stderr)
            #print(process.stderr, file=sys.stderr)
            #print("test:", process, file=sys.stderr)

            # Use ASE to read back in the .castep file
            castep_ase.read(f"{SEEDNAME}.castep")
            #def do_nothing(*args, **kwargs):
            #    pass

            #castep_ase.update = do_nothing

            energy = castep_ase._energy_free * units.eV
            forces = np.array(castep_ase._forces) * units.eV / units.angstrom

            # Set results
            RESULTS = Results(energy, forces)


        # Broadcast results to all nodes and return
        mpi_comm.bcast(RESULTS)
        return RESULTS

    @staticmethod
    @requires_env(CastepDebugEnv)
    def get_results(mpi_comm):
        return RESULTS



# ASE interface -- Load if possible
try:
    from .calculator.dft_code_calculator import DFT_Calculator
except ModuleNotFoundError:
    raise #pass

# Utils for return values
from .utils import units
from .utils.results import Results

# Tools for launching master and worker threads
from .master.master_io import Worker

# Tasks to expose on top level
from .tasks.castep_debug_task import CastepDebugTask
from .tasks.castep_task import CastepTask
from .tasks.task import Task


"""IO operations for worker.pyp"""
import sys
import pickle
import os

class ParentIO:
    """Class that handles receiving and communicating data to the parent python instance"""
    def __init__(self, comm):
        """Initialises a reader for io operations to the master process"""
        self.comm = comm
        self.rank = comm.Get_rank()
        self.isroot = self.rank == 0

        # Open named pipe
        if self.isroot:
            self.fname = self._recv()
            self.pipe = open(self.fname, 'wb')

    def _recv(self):
        """Read pickled object from stdin only on the root node"""
        if self.isroot:
            data = pickle.load(sys.stdin.buffer)
        else:
            data = None
        return data

    def recv_all(self):
        """Read pickled object from stdin and broadcast to all nodes"""
        return self.comm.bcast(self._recv())

    def send(self, data):
        """Send data to master process from root node"""
        if self.isroot:
            pickle.dump(data, self.pipe)
            self.pipe.flush()



"""Module defining a Task and a wrapper to say what environment we want to run it in."""
from functools import wraps

def requires_env(envtype: type):
    """Wrapper that forces the required environment to be loaded"""
    def env_wrapper(func):
        @wraps(func)
        def wrapped_func(environments, mpi_comm, *args, **kwargs):
            environments.get_env(envtype, mpi_comm) # Needs comm for barrier sync
            return func(mpi_comm, *args, **kwargs)
        return wrapped_func
    return env_wrapper

class Task:
    """This is a class to act as a task for the worker to do."""
    def __init__(self, task_fn, *args, **kwargs):
        """Creates a task object that can be created on the master and executed by the workers"""
        self.task = task_fn
        self.args = args
        self.kwargs = kwargs
        self.results = None

    def run(self, environments, mpi_comm):
        """Run the task (to be called on the worker)"""
        self.results = self.task(environments, mpi_comm, *self.args, **self.kwargs)

    def __repr__(self):
        repr_str = f"TaskType: {type(self).__name__} Task: {self.task.__name__}"
        if self.args:    
            repr_str += f"\n\tDataArgs: {self.args}"
        if self.kwargs:  
            repr_str += f"\n\tDataKWargs: {self.kwargs}"
        return repr_str



#!/usr/bin/env python3
"""Main entry point for project"""
from ..tasks.task import Task
from ..master.master_io import Worker
from ..utils import units
import numpy as np
import functools
from inspect import getcallargs
from ase.calculators.calculator import Calculator, PropertyNotImplementedError, compare_atoms
from ase.units import Bohr, Hartree
from time import sleep

# Fancy print function call and args
def debug_print_entry(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        if self.debug:
            output = f"Called {func.__name__}"
            fancy_kwargs = getcallargs(func, self, *args, **kwargs)
            arg_str = "(" + ",".join(f"{k}={v}" for k,v in fancy_kwargs.items() if k != "self") + ")"
            print(output + arg_str)
        return func(self, *args, **kwargs)
    return wrapped

# Conversion helper from pint to ase units. The exact units don't matter so 
# long as they correspond to the same thing in both libraries
ase_units = {
    'energy': (units.hartree, Hartree),
    'free_energy': (units.hartree, Hartree),
    'forces': (units.hartree/units.bohr, Hartree/Bohr),
    'stress': (units.hartree/units.bohr**3, Hartree/Bohr**3)
}

class DFT_Calculator(Calculator):
    implemented_properties = ['energy','forces','stress','free_energy']
    ignored_changes = {}
    discard_results_on_any_change = True

    def __init__(self, atoms=None, label='run', calculator_type=None, debug=False, num_mpi_processes=1,
                **kwargs):
        
        self.worker_started = False
        self.initialised = False
        self.reset_needed = False
        self.num_processes = num_mpi_processes
        self.worker = None
        self.debug = debug

        assert issubclass(calculator_type, Task), "Must pass a DFT calculator task type!"
        self.dft = calculator_type

        self.default_parameters = self.dft.get_default_parameters('energy', atoms)
        self.full_reset_flags = {'numbers'}

        super().__init__(atoms=atoms, label=label, **kwargs)
        
        self._start_dft()
        
    @debug_print_entry
    def _restart_dft(self):
        """Restart the DFT subprocess if required (ie if it has been initialised)"""
        if self.initialised:
            self._stop_dft()
            # print("resetting dft process")
        if not self.worker_started:
            self._start_dft()
        self.reset_needed = False


    def _start_dft(self):
        """Start the DFT code backend"""
        if self.worker_started:
            raise Exception("Tried to initialise an already initialised dft code!")

        # Start a worker manager and spawn a worker
        self.worker = Worker(self.label, self.num_processes)

        # Set flags to say this has been started but not initialised
        self.worker_started = True
        self.initialised = False

    def _stop_dft(self):
        """Stops the DFT code backend"""
        if not self.worker_started:
            raise Exception("Tried to end a worker that was already stopped!")

        # Send DFT finalisation task
        self.worker.add_task(0, self.dft(self.dft.finalise))
        self.worker.get_next_finished()
        self.worker.finish()

        # Give a small grace period and then kill it
        sleep(0.1)
        self.worker.kill()

        # Reset status flags
        self.worker = None
        self.worker_started = False
        self.initialised = False

    def set(self, **kwargs):
        # Since we can't change any parameters right now, we always restart
        changed_params = super().set(**kwargs)
        if changed_params:
            #  Mark for reset later on
            self.reset_needed = True

    @debug_print_entry
    def calculate(self, atoms=None, properties=None, system_changes=None):
        """Call the DFT code to calculate the requested properties"""

        # Check if we need to do a full reset
        if self.reset_needed or any(flag in system_changes for flag in self.full_reset_flags):
            self._restart_dft()

        # Make sure the worker is actually started
        if not self.worker_started:
            self._start_dft()

        # Make a copy of atoms
        if atoms is not None:
            self.atoms = atoms.copy()

        # Either initialise with these atoms or update these atoms
        if not self.initialised:
            self.worker.add_task(1, self.dft(self.dft.init, self.atoms, self.parameters))
            self.initialised = True
        elif system_changes:
            self.worker.add_task(1, self.dft(self.dft.update_atoms, self.atoms))

        # Do the ground state calculation
        want_forces = 'forces' in properties
        want_stress = 'stress' in properties
        self.worker.add_task(10, self.dft(self.dft.calculate, want_forces, want_stress))

        # ...and wait for the result
        task_id, result = self.worker.get_next_result()
        assert task_id == 10, "Wrong task number returned from DFT worker process"

        # Convert units to ASE and store in results
        for key,value in result.__dict__.items():
            if value is not None:
                try:
                    pint_unit, ase_unit = ase_units[key]
                except KeyError as exc:
                    raise Exception(f"{key} does not have a unit conversion") from exc
                
                self.results[key] = value.to(pint_unit).magnitude * ase_unit


    




"""Module defining control flags that can be sent to the worker"""
from enum import Enum, auto, unique

@unique
class ControlFlag(Enum):
    """Defines some control flags that can be sent to/from the worker"""
    CONTINUE = auto()
    END = auto()
    CRASH = auto()
    WORKER_CRASHED = auto()



"""Defines a global set of units we can use to wrap outputs from DFT code"""
import pint

ureg = pint.UnitRegistry()
pint.set_application_registry(ureg)

hartree = ureg.sys.atomic.hartree
bohr = ureg.sys.atomic.bohr

eV = ureg.sys.SI.electron_volt
angstrom = ureg.sys.SI.angstrom

def mag(x, unit_value):
    return x.to(unit_value).magnitude


#!/usr/bin/env python3
"""Main worker script that is run in parallel"""
import sys
import os
from mpi4py import MPI
from dft_python_api.utils.controlflags import ControlFlag
from dft_python_api.worker.worker_io import ParentIO
from dft_python_api.tasks.worker_environment import Envs
from dft_python_api.tasks.task import Task


def get_printfn(worker_id, rank, debug=True):
    """Returns a stderr print function with tags for process information"""
    def printfn(*args, **kw):
        """
        Print function that writes to stderr rather than stdout,
        as this is connected to the master process
        """
        print(f"<Worker {worker_id}.{rank}> ", end='')
        print(*args, **kw)
        sys.stdout.flush()

    def noprint(*args, **kw):
        pass

    if debug:
        return printfn
    else:
        return noprint


def worker_main():
    """
    This is the main loop for a parallel worker.
    """
    # Get some MPI parameters
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    on_root = rank == 0

    # Create a reader/writer to the main process
    parent_io = ParentIO(comm)

    # Read worker id and debug level
    worker_id = parent_io.recv_all()

    # Generate pretty printer
    printfn = get_printfn(worker_id, rank, debug=False)

    # Create empty environments set
    envs = Envs(worker_id, on_root)

    while True:

        # Receive data from master
        task = parent_io.recv_all()
        results = None

        # Check if the data is an end signal
        if task == ControlFlag.END:
            break
        if task is ControlFlag.CRASH:
            if on_root:
                printfn("Crashing on command....")
                comm.Abort()
        elif isinstance(task, Task):
            printfn(f'Running task: {task}')
            task.run(envs, comm)
            results = task.results
            if results:
                printfn(f"Finished. Results: {results}")
            else:
                printfn(f"Finished.")
        else:
            printfn("Unknown object recieved?")
            comm.Abort()


        #Send data back to master
        parent_io.send(results)


if __name__ == "__main__":
    worker_main()



"""Module containing CASTEP specific task information"""
#pylint: disable=import-outside-toplevel, no-self-argument
from pathlib import Path
import sys
import numpy as np
import ase.io
from ..utils import units
from ..tasks.task import Task, requires_env
from ..tasks.worker_environment import Env
from ..utils.results import Results
import glob

ATOMS = {}
PARAMS = {}
RESULTS = {}

SEEDNAME = 'run'

class CastepEnv(Env):
    """Get a folder to run this instance of CASTEP in"""
    def setup(self):
        self.id_name = "castep"


def write_dict(filename, options):
    """Write options to file"""
    with open(filename, "w") as file_handle:
        for key, val in options.items():
            file_handle.write(f"{key} : {val}\n")


class write_helper():
    def __init__(self, atoms):
        self.atoms = atoms


    # Copied from ASE Castep calculator -- required for cell writing
    def _get_number_in_species(self, at, atoms=None):
        """Return the number of the atoms within the set of it own
        species. If you are an ASE commiter: why not move this into
        ase.atoms.Atoms ?"""
        if atoms is None:
            atoms = self.atoms
        numbers = atoms.get_atomic_numbers()
        n = numbers[at]
        nis = numbers.tolist()[:at + 1].count(n)
        return nis

    def _get_absolute_number(self, species, nic, atoms=None):
        """This is the inverse function to _get_number in species."""
        if atoms is None:
            atoms = self.atoms
        ch = atoms.get_chemical_symbols()
        ch.reverse()
        total_nr = 0
        assert nic > 0, 'Number in species needs to be 1 or larger'
        while True:
            if ch.pop() == species:
                if nic == 1:
                    return total_nr
                nic -= 1
            total_nr += 1

class CastepTask(Task):
    """CASTEP specific tasks"""

    @staticmethod
    def get_error(worker_id):
        """Try and get the .err file"""
        err_files = glob.glob(f"./work/worker-{worker_id}-castep/*.err")
        if not err_files:
            return "No error file available"
        with open(err_files[0]) as err_file:
            err_msg = err_file.read()
        return err_msg

    @staticmethod
    def get_default_parameters(task, atoms=None, **extra_args):
        """
        Sets the default parameters object that we want. Override arguments with extra_args.
        This function is special and does not require a communicator and should be called by
        the main script.
        TODO: Maybe add a quality setting?
        """
        # Set the default parameters
        params = {
            "write_checkpoint": "None",
            "cutoff_energy": "200 eV",
            "opt_strategy": "speed"
        }

        if task in ["energy", "singlepoint", "SPE", "forces"]:
            params["task"] = "SPE"
        elif task in ['geometry']:
            params["task"] = "geometryoptimisation"

        else:
            raise ValueError("Invalid value for task: {}".format(task))

        # Add extra ones passed in
        params.update(extra_args)

        #Return the parameters object
        return params

    @staticmethod
    @requires_env(CastepEnv)
    def init(mpi_comm, atoms, params):
        """Initialise the DFT code"""
        from castep_python import Castep_Python
        global ATOMS, PARAMS
        ATOMS = atoms
        PARAMS = params

        # Write cell and param file for now
        param_file = f"{SEEDNAME}.param"
        cell_file = f"{SEEDNAME}.cell"

        # Write cell with ASE
        atoms.calc = write_helper(atoms)  # Hack to make cell contraints work
        ase.io.write(cell_file, ATOMS)

        # Write params file
        write_dict(param_file, PARAMS)

        # Do castep internal init
        Castep_Python.initialise(mpi_comm.py2f(), SEEDNAME)


    @staticmethod
    @requires_env(CastepEnv)
    def finalise(mpi_comm):
        """Delete any static variables -- also call code cleanup methods here"""
        from castep_python import Castep_Python
        global ATOMS, PARAMS, RESULTS
        ATOMS = {}
        PARAMS = {}
        RESULTS = {}

        #Castep_Python.finalise()

    @staticmethod
    @requires_env(CastepEnv)
    def update_params(mpi_comm, new_params):
        """Update the parameters set"""
        from castep_python import Castep_Python
        # Do some validation checking here
        pass

        # Now update the saved params
        PARAMS.update(new_params)

        # Now pass these new params to the code
        for key, value in new_params.items():
            Castep_Python.update_param(key, value)

    @staticmethod
    @requires_env(CastepEnv)
    def update_atoms(mpi_comm, new_atoms):
        """Update the atoms list"""
        from castep_python import Castep_Python
        global ATOMS

        # Optional validation checking
        if (new_atoms.get_atomic_numbers() != ATOMS.get_atomic_numbers()).any():
            raise NotImplementedError("Castep does not support changing atomic species")

        # Attach units to pos and lattice
        pos = new_atoms.get_positions().T * units.angstrom 
        lattice = new_atoms.get_cell() * units.angstrom   

        # # Update lattice if that changed

        if (new_atoms.get_cell() != ATOMS.get_cell()).any():
            Castep_Python.update_lattice_and_positions(lattice.to(units.bohr).magnitude, pos.to(units.bohr).magnitude)
        else:
            # Otherwise just update the positions
            Castep_Python.update_positions(pos.to(units.bohr).magnitude)

        # Save current atoms
        ATOMS = new_atoms


    @staticmethod
    @requires_env(CastepEnv)
    def update_atoms_extrap(mpi_comm, new_atoms):
        """Update the atoms list"""
        from castep_python import Castep_Python
        global ATOMS

        # Optional validation checking
        if (new_atoms.get_atomic_numbers() != ATOMS.get_atomic_numbers()).any():
            raise NotImplementedError("Castep does not support changing atomic species")

        pos = new_atoms.get_positions().T * units.angstrom

        # Now update saved atoms
        Castep_Python.update_positions_extrap(pos.to(units.bohr).magnitude)
        ATOMS = new_atoms

        
    @staticmethod
    @requires_env(CastepEnv)
    def calculate(mpi_comm, want_forces=False, want_stress=False):
        """Do the actual calculation"""
        from castep_python import Castep_Python
        global RESULTS

        num_atoms = len(ATOMS.get_atomic_numbers())

        # Call re-entrant castep method
        energy = Castep_Python.calculate_energy()

        # Set results
        RESULTS = Results(energy * units.hartree)
        RESULTS.free_energy = RESULTS.energy

        if want_forces:
            forces = np.empty((3, num_atoms), order='F')
            Castep_Python.calculate_forces(forces)
            RESULTS.forces = forces.T * units.hartree / units.bohr

        if want_stress:
            stress = np.empty(6, order='F')
            Castep_Python.calculate_stress(stress)
            RESULTS.stress = stress * units.hartree / units.bohr**3

        # Broadcast results to all nodes and return
        mpi_comm.bcast(RESULTS)
        return RESULTS

    @staticmethod
    @requires_env(CastepEnv)
    def get_results(mpi_comm):
        return RESULTS



"""Module containing CASTEP specific task information"""
#pylint: disable=import-outside-toplevel, no-self-argument
from pathlib import Path
import sys
import numpy as np
import ase.io
from ..tasks.task import Task, requires_env
from ..tasks.worker_environment import Env
from ..utils.results import Results
from ..utils import units

# Make sure fortran files can be found... hacky
file = Path(__file__).resolve()
fortran_dir = file.parent / 'fortran_test_example'
sys.path.append(str(fortran_dir))


class FortEnv(Env):
    """Get a folder to run this instance in"""
    def setup(self):
        self.id_name = "fortran_test"

class FortTask(Task):
    """Fortran task example linking to code in fortan_test_example"""

    @staticmethod
    @requires_env(FortEnv)
    def add(mpi_comm, a, b):

        # Import the fortran module "fortran_test"
        # from the library file "fortran_code"
        from fortran_code import fortran_test # type: ignore

        # Fortran library call
        added_values = fortran_test.add_fort(a, b)

        # Return the result
        return added_values

    @staticmethod
    @requires_env(FortEnv)
    def comm_info(mpi_comm):
        # Import the fortran module "fortran_test"
        # from the library file "fortran_code"
        from fortran_code import fortran_test # type: ignore


        # Fortran library call (using mpi comm handle for fortran)
        comm_size = fortran_test.comm_info_fort(mpi_comm.py2f())

        # Return the result
        return comm_size
        



"""Classes to control communication with parallel subprocesses"""
import os
import sys
import signal
import psutil
import pickle
import time
import atexit
import threading
from subprocess import Popen, PIPE
from pathlib import Path
from queue import Queue
from ..utils.controlflags import ControlFlag

class WorkerDiedException(Exception):
    pass

class WorkerCrashed(Exception):
    """Indicates that a worker has crashed"""


def get_default_worker_script():
    """Get the default worker.py script"""
    # import dft_python_api.worker as worker_init_script
    # root_py = Path(os.path.realpath(worker_init_script.__file__))
    # return root_py.parent / 'worker.py'
    return "-mdft_python_api.worker.worker"

# Cleanup code
DEBUG_CLEANUP=False

def cleanup_process(process: psutil.Process):
    """Terminate process and all children (ie mpirun children)"""
    if DEBUG_CLEANUP:
        print("killing process", process)

    # Gather child pids to kill
    try:
        children = list(process.children(recursive=True))
    except psutil.NoSuchProcess:
        children = []

    try:
        process.kill()
    except psutil.NoSuchProcess:
        pass

    # Kill children (ie worker threads that may have a different pgid )
    for c in children:
        if DEBUG_CLEANUP:
            print("killing child", c)
        c.kill()
    
def cleanup_file(pipe_file):
    """Delete a file"""
    try:
        if DEBUG_CLEANUP:
            print("Removing pipe", pipe_file)
        os.remove(pipe_file)
    except FileNotFoundError:
        pass # Ignore if already deleted

class Worker:
    """Represents a collection of workers of variable sizes"""
    def __init__(self, name, max_cores, single_worker_mode=True, mpirun_command="mpirun", mpirun_flags=None):
        self.name = name
        self.max_cores = max_cores
        self.used_cores = 0
        self.current_worker_id = 0
        self.workers = dict()

        self.mpirun_command = mpirun_command
        if mpirun_flags is None:
            self.mpirun_flags = ["-np", "{num_cores}"]
        else:
            self.mpirun_flags = mpirun_flags

        self.work_queue = Queue()
        self.results_queue = Queue()

        if single_worker_mode:
            self.add_workers(max_cores)

    def add_workers(self, num_cores_per_worker, num_workers=1):
        """Add more workers to the worker pool"""
        if self.used_cores + num_workers * num_cores_per_worker > self.max_cores:
            raise ValueError("Attempted to spawn more worker threads than max_cores!")
        self.used_cores += num_workers * num_cores_per_worker

        for _ in range(num_workers):
            worker_id = f"{self.name}-{self.get_next_id()}"
            worker = WorkerIO(worker_id=worker_id,
                       num_cores=num_cores_per_worker,
                       work_queue=self.work_queue,
                       results_queue=self.results_queue,
                       mpirun_command=self.mpirun_command,
                       mpirun_flags=self.mpirun_flags
            )
            self.workers[worker_id] = worker

    def add_task(self, task_id, task):
        """Adds a task to the queue"""
        self.work_queue.put((task_id, task))

    def get_next_id(self):
        """Get the next free worker id"""
        self.current_worker_id += 1
        return self.current_worker_id


    def kill(self):
        """Kills all subprocesses"""
        for worker in self.workers.values():
            worker.kill()

    def finish(self):
        """Sends the polite finish message to all workers"""
        assert self.work_queue.empty(), "Work queue wasn't empty!"
        for worker in self.workers.values():
            worker.finish()

    def get_next_finished(self):
        """Return the next finished task"""
        return_value = self.results_queue.get()

        task_id, result, *error_info = return_value

        if result == ControlFlag.WORKER_CRASHED:
            error_task, worker_id = error_info
            if hasattr(error_task, "get_error"):
                err_msg = error_task.get_error(worker_id)
                raise WorkerDiedException(f"Worker died on task with message:\n{error_task}\n{err_msg}")
            else:
                raise WorkerDiedException(f"Worker died on task {task_id}")
        else:
            return task_id, result


    def get_next_result(self):
        """Return the next finished task which returns a result"""
        result = None
        while result is None:
            task_id, result = self.get_next_finished()
        return task_id, result

class WorkerIO:
    """Represents an individual worker"""
    def __init__(self, worker_id, num_cores, work_queue, results_queue, mpirun_command, mpirun_flags, worker_py=None):
        """Initialise a worker thread"""
        if worker_py is None:
            worker_py = get_default_worker_script()

        # Use current python executable
        py_exe = sys.executable

        self.num_cores = num_cores

        # We replace named entries in flags -- eg {num_cores} becomes the number of cores
        # TODO: Test this with other mpirun style things (srun etc)
        replacements = {
            "num_cores": str(num_cores)
            }

        mpi_flags = [f.format(**replacements) for f in mpirun_flags]

        self.process_cmd = mpirun_command, *mpi_flags, py_exe, str(worker_py)

        self.worker_id = worker_id
        self.work_queue = work_queue
        self.results_queue = results_queue

        # Start the process
        self.process = Popen(self.process_cmd, shell=False,
                             stdin=PIPE, bufsize=0)   # PIPE with no buffering

        # Immediately add to kill at exit list
        atexit.register(cleanup_process, psutil.Process(self.process.pid))

        # Make a named pipe for this process
        self.fname = f'.worker-{id(self)}-iopipe'
        os.mkfifo(self.fname, mode=0o777)
        atexit.register(cleanup_file, self.fname)
        
        # Send worker id and pipe name
        self.send(self.fname)           # Send pipe name
        self.send(self.worker_id)       # Send worker id

        # Now open pipe for reading
        self.pipe = open(self.fname, 'rb')

        # Go into work loop
        self.thread = threading.Thread(target=self.work_loop,
                                       name=f"worker-{worker_id}",
                                       daemon=True)
        self.thread.start()

    def work_loop(self):
        """Take items from work queue, process them and send results to result_queue"""
        while True:
            task_id, task = self.work_queue.get()
            try:
                self.send(task)
            except BrokenPipeError:
                print(f"Died sending data: {task_id, task}")
                self.results_queue.put((task_id, ControlFlag.WORKER_CRASHED, task, self.worker_id))
                break

            if task == ControlFlag.END:
                break

            try:
                result = self.read()
            except (BrokenPipeError, EOFError):
                print(f"Died while waiting on response for task: {task_id, task}")
                self.results_queue.put((task_id, ControlFlag.WORKER_CRASHED, task, self.worker_id))
                break

            self.results_queue.put((task_id, result))
            self.work_queue.task_done()



    def send(self, data: object):
        """Send object to child process"""
        pickle.dump(data, self.process.stdin)
        self.process.stdin.flush()

    def read(self) -> object:
        """Read object from child process synchronously"""
        return pickle.load(self.pipe)


    def __repr__(self):
        return "Worker(id={0.worker_id}, num_cores={0.num_cores})".format(self)


    def finish(self):
        """Polite message to thread and subprocess to close"""
        self.work_queue.put((-1, ControlFlag.END))

    def kill(self):
        """Send kill signal to subprocess"""

        # First enumerate children
        parent = psutil.Process(self.process.pid)
        children = list(parent.children(recursive=True))

        # Send kill to mpirun
        self.process.kill()

        # Kill children (ie worker threads that may have a different pgid )
        for c in children:
            c.kill()

        time.sleep(0.1)

        try:
            os.remove(self.fname)
        except FileNotFoundError:
            pass # Ignore if already deleted
